from copy import deepcopy
from typing import Generator, Optional, Tuple, Union, Dict, List, Any
from itertools import combinations
import pandas as pd
import numpy as np
from collections import ChainMap, defaultdict
from joblib import Parallel, delayed
from math import comb
import warnings
from sklearn.exceptions import ConvergenceWarning

from .purged_kfold import PurgedKFold

class CombinatorialPurged(PurgedKFold):
    """
    Combinatorial Purged Cross-Validation (CPCV) implementation based on Marcos Lopez de Prado's method.

    This class provides a cross-validation scheme that aims to address the main drawback of the Walk Forward
    and traditional Cross-Validation methods by testing multiple paths. Given a number of backtest paths,
    CPCV generates the precise number of combinations of training/testing sets needed to generate those paths,
    while purging training observations that might contain leaked information.

    Parameters
    ----------
    n_splits : int
        Number of splits/groups to partition the data into.
    n_test_groups : int
        Size of the testing set in terms of groups.
    times : Union[pd.Series, Dict[str, pd.Series]]
        The timestamp series associated with the labels.
    embargo : float
        The embargo rate for purging.
    """


    @staticmethod
    def _path_locations(
        n_splits: int,
        combinations_: List[Tuple[int]]
    ) -> Dict[int, List[Tuple[int, int]]]:
        """
        Generate a labeled path matrix and return path locations for N choose K.

        This method generates a matrix where each entry corresponds to a specific combination of
        training/testing sets, and helps in mapping these combinations to specific backtest paths.

        Parameters
        ----------
        n_splits : int
            Number of splits/groups to partition the data into.
        combinations_ : list
            List of combinations for training/testing sets.

        Returns
        -------
        dict
            A dictionary mapping each backtest path to its corresponding train/test combination.
        """

        # Initialize a zero matrix
        matrix = np.zeros((n_splits, len(combinations_)), dtype=int)

        # Set appropriate entries to 1 and label path numbers in the same loop
        for col, indices in enumerate(combinations_):
            matrix[indices, col] = 1

        # Label path numbers
        def label_path(row):
            counter = iter(range(1, row.sum() + 1))
            return [next(counter) if x == 1 else 0 for x in row]

        path_numbers = np.array(list(map(label_path, matrix)))

        # Extract path locations using loops
        def map_to_location(coord):
            i, j = coord
            value = path_numbers[i, j]
            locations[value].append((i, j))

        rows, cols = path_numbers.shape
        locations = defaultdict(list)
        list(map(map_to_location, [(x, y) for x in range(rows) for y in range(cols) if path_numbers[x, y] != 0]))

        return dict(locations)


    @staticmethod
    def _combinatorial_splits(
        combinations_: List[Tuple[int]],
        split_segments: np.ndarray
    ) -> Generator[np.ndarray, None, None]:
        """
        Generate combinatorial test sets based on the number of test groups (n_test_groups).

        This method creates test sets by considering all possible combinations of group splits, allowing
        for the creation of multiple test paths, as described in the CPCV methodology.

        Parameters
        ----------
        combinations_ : list
            List of combinations for training/testing sets.
        split_segments : np.ndarray
            Array of data split segments.

        Returns
        -------
        Generator[np.ndarray]
            A generator yielding the combinatorial test sets.
        """

        for test_groups in combinations_:
            test_sets = [split for i, split in enumerate(split_segments) if i in test_groups]
            yield np.concatenate(test_sets)


    def __init__(
        self,
        n_splits: int,
        n_test_groups: int,
        times: Union[pd.Series, Dict[str, pd.Series]],
        embargo: float = 0
    ):
        """
        Initialize the CombinatorialPurged class.

        Parameters
        ----------
        n_splits : int
            Number of splits/groups to partition the data into.
        n_test_groups : int
            Size of the testing set in terms of groups.
        times : Union[pd.Series, Dict[str, pd.Series]]
            The timestamp series associated with the labels.
        embargo : float
            The embargo rate for purging.
        """

        super().__init__(n_splits, times, embargo)
        self.n_test_groups = n_test_groups

    def get_n_splits(
        self,
        data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]] = None,
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> int:
        """
        Return number of splits.

        :param data: Dataset or dictionary of datasets.
        :type data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]]

        :param labels: Labels or dictionary of labels.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]

        :param groups: Group labels for the samples.
        :type groups: Optional[np.ndarray]

        :return: Number of splits.
        :rtype: int
        """
        return comb(self.n_splits, self.n_test_groups)

    def _single_split(
        self,
        single_times: np.ndarray,
        single_data: np.ndarray,
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Splits data into train and test indices based on the defined combinatorial splits.

        This function is used to generate multiple train-test splits based on the combinatorial
        cross-validation method. It ensures that each train-test split is properly purged and
        embargoed to prevent data leakage.

        :param single_times: Timestamp series associated with the labels.
        :param single_data: The input data to be split.

        :return: Generator that yields tuples of (train indices, test indices).

        .. note:: The function validates the input, and uses combinatorial cross-validation method to
                produce the train-test splits.
        """

        self._validate_input(single_times, single_data)

        indices = np.arange(single_data.shape[0])
        split_segments = np.array_split(indices, self.n_splits)
        combinations_ = list(combinations(range(self.n_splits), self.n_test_groups))

        all_combinatorial_splits = list(
            CombinatorialPurged._combinatorial_splits(combinations_, split_segments)
        )

        for test_indices in all_combinatorial_splits:
            train_indices = self._get_train_indices(test_indices, single_times)
            yield train_indices, test_indices


    def _combinations_and_path_locations_and_split_segments(
        self,
        data: pd.DataFrame
    ) -> Tuple[List[Tuple[int, ...]], Dict[str, np.ndarray], List[np.ndarray]]:
        """
        Generate combinations, path locations, and split segments for the data.

        This function is a helper that computes necessary components for combinatorial cross-validation.

        :param data: The input dataframe to generate combinations, path locations, and split segments.

        :return: Tuple containing combinations, path locations, and split segments.

        .. math::
        \\text{combinations} = \\binom{n}{k}
        """

        combinations_ = list(combinations(range(self.n_splits), self.n_test_groups))
        locations = CombinatorialPurged._path_locations(self.n_splits, combinations_)

        indices = np.arange(data.shape[0])
        split_segments = np.array_split(indices, self.n_splits)

        return combinations_, locations, split_segments


    def _single_backtest_paths(
            self,
            single_times: pd.Series,
            single_data: pd.DataFrame,
    ) -> Dict[str, List[Dict[str, List[np.ndarray]]]]:
        """
        Generate the backtest paths for given input data.

        This function creates multiple backtest paths based on combinatorial splits, where
        each path represents a sequence of train-test splits. It ensures that data leakage
        is prevented by purging and applying embargo to the train-test splits.

        :param single_times: Timestamp series associated with the data.
        :param single_data: Input data on which the backtest paths are based.

        :return: A dictionary where each key is a backtest path name, and the value is
                a list of dictionaries with train and test index arrays.

        .. note:: This function relies on combinatorial cross-validation for backtesting to
                generate multiple paths of train-test splits.
        """

        self._validate_input(single_times, single_data)

        paths = {}
        combinations_, locations, split_segments = self._combinations_and_path_locations_and_split_segments(single_data)
        all_combinatorial_splits = list(
            CombinatorialPurged._combinatorial_splits(combinations_, split_segments)
        )

        for path_num, locs in locations.items():
            path_data = []

            for (G, S) in locs:
                # Use all_combinatorial_splits[S] to determine potential training indices
                # Now, we derive the actual training set after purge and embargo
                train_indices = self._get_train_indices(all_combinatorial_splits[S], single_times)

                path_data.append({
                    "Train": np.array(train_indices),
                    "Test": split_segments[G],
                })

            paths[f"Path {path_num}"] = np.array(path_data)

        return paths


    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: np.ndarray,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Generate predictions for a single backtest using combinatorial splits.

        This method calculates predictions across various paths created by combinatorial splits
        of the data. For each combinatorial split, a separate estimator is trained and then used
        to predict on the corresponding test set.

        :param single_estimator: The machine learning model or estimator to be trained.
        :param single_times: Timestamps corresponding to the data points.
        :param single_data: Input data on which the model is trained and predictions are made.
        :param single_labels: Labels corresponding to the input data.
        :param single_weights: Weights for each data point.
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: Number of CPU cores to use for parallelization. Default is 1.

        :return: A dictionary where keys are path names and values are arrays of predictions.

        .. note:: This function relies on internal methods (e.g., `_get_train_indices`)
                to manage data splits and training.

        .. note:: Parallelization is used to speed up the training of models for different splits.
        """

        self._validate_input(single_times, single_data)

        if single_weights is None:
            single_weights = np.ones((len(single_data),))

        paths_predictions = {}
        combinations_, locations, split_segments = self._combinations_and_path_locations_and_split_segments(single_data)

        def train_single_estimator(
            single_estimator_,
            combinatorial_test_indices
        ):
            combinatorial_train_indices = self._get_train_indices(combinatorial_test_indices, single_times)

            X_train = single_data.iloc[combinatorial_train_indices]
            y_train = single_labels.iloc[combinatorial_train_indices]
            weights_train = single_weights[combinatorial_train_indices]

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    single_estimator_.fit(X_train, y_train, sample_weight=weights_train)

                except TypeError:
                    single_estimator_.fit(X_train, y_train)

            return single_estimator_

        combinatorial_trained_estimators = Parallel(n_jobs=n_jobs)(
            delayed(train_single_estimator)(deepcopy(single_estimator), combinatorial_test_indices)
            for combinatorial_test_indices in CombinatorialPurged._combinatorial_splits(combinations_, split_segments))


        def get_path_data(
            path_num,
            locs
        ):
            path_data = []

            for (G, S) in locs:
                test_indices = split_segments[G]
                X_test = single_data.iloc[test_indices]

                if predict_probability:
                    predictions = combinatorial_trained_estimators[S].predict_proba(X_test)

                else:
                    predictions = combinatorial_trained_estimators[S].predict(X_test)

                path_data.extend(predictions)

            path_data = {f"Path {path_num}" : np.array(path_data)}

            return path_data

        paths_predictions = Parallel(n_jobs=n_jobs)(delayed(get_path_data)(path_num, locs) for path_num, locs in locations.items())
        paths_predictions = dict(ChainMap(*reversed(paths_predictions)))

        return paths_predictions



import numpy as np
import scipy.stats as ss
import math
from sklearn.metrics import mutual_info_score


def calculate_variation_of_information(
    x: np.ndarray,
    y: np.ndarray,
    bins: int,
    norm: bool = False
) -> float:
    """
    Calculates Variation of Information.

    :param x: First data array.
    :param y: Second data array.
    :param bins: Number of bins for the histogram.
    :param norm: If True, the result will be normalized.

    :return: Variation of Information.
    
    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    Methodology: Snippet 3.2, Page 44
    """
    histogram_xy = np.histogram2d(x, y, bins)[0]
    mutual_information = mutual_info_score(None, None, contingency=histogram_xy)
    marginal_x = ss.entropy(np.histogram(x, bins)[0])
    marginal_y = ss.entropy(np.histogram(y, bins)[0])
    variation_xy = marginal_x + marginal_y - 2 * mutual_information

    if norm:
        joint_xy = marginal_x + marginal_y - mutual_information
        variation_xy /= joint_xy

    return variation_xy


def calculate_number_of_bins(
    num_observations: int,
    correlation: float = None
) -> int:
    """
    Calculates the optimal number of bins for discretization.

    :param num_observations: Number of observations.
    :param correlation: Correlation value. If None, the function will use the univariate case.

    :return: Optimal number of bins.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    Methodology: Snippet 3.3, Page 46
    """
    if correlation is None:
        z = (8 + 324 * num_observations + 12 * (36 * num_observations + 729 * num_observations ** 2) ** .5) ** (1 / 3.)
        bins = round(z / 6. + 2. / (3 * z) + 1. / 3)
    else:
        bins = round(2 ** -.5 * (1 + (1 + 24 * num_observations / (1. - correlation ** 2)) ** .5) ** .5)
    return int(bins)


def calculate_variation_of_information_extended(
    x: np.ndarray,
    y: np.ndarray,
    norm: bool = False
) -> float:
    """
    Calculates Variation of Information with calculating number of bins.

    :param x: First data array.
    :param y: Second data array.
    :param norm: If True, the result will be normalized.

    :return: Variation of Information.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    Methodology: Snippet 3.3, Page 46
    """
    num_bins = calculate_number_of_bins(x.shape[0], correlation=np.corrcoef(x, y)[0, 1])
    histogram_xy = np.histogram2d(x, y, num_bins)[0]
    mutual_information = mutual_info_score(None, None, contingency=histogram_xy)
    marginal_x = ss.entropy(np.histogram(x, num_bins)[0])
    marginal_y = ss.entropy(np.histogram(y, num_bins)[0])
    variation_xy = marginal_x + marginal_y - 2 * mutual_information

    if norm:
        joint_xy = marginal_x + marginal_y - mutual_information
        variation_xy /= joint_xy

    return variation_xy


def calculate_mutual_information(
    x: np.ndarray,
    y: np.ndarray,
    norm: bool = False
) -> float:
    """
    Calculates Mutual Information with calculating number of bins.

    :param x: First data array.
    :param y: Second data array.
    :param norm: If True, the result will be normalized.

    :return: Mutual Information.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    Methodology: Snippet 3.4, Page 48
    """
    num_bins = calculate_number_of_bins(x.shape[0], correlation=np.corrcoef(x, y)[0, 1])
    histogram_xy = np.histogram2d(x, y, num_bins)[0]
    mutual_information = mutual_info_score(None, None, contingency=histogram_xy)

    if norm:
        marginal_x = ss.entropy(np.histogram(x, num_bins)[0])
        marginal_y = ss.entropy(np.histogram(y, num_bins)[0])
        mutual_information /= min(marginal_x, marginal_y)

    return mutual_information


def calculate_distance(
    dependence: np.ndarray,
    metric: str = "angular"
) -> np.ndarray:
    """
    Calculates distance from a dependence matrix.

    :param dependence: Dependence matrix.
    :param metric: Metric used to calculate distance. Available options are "angular" and "absolute_angular".

    :return: Distance matrix.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    """
    if metric == "angular":
        distance = ((1 - dependence).round(5) / 2.) ** 0.5
    elif metric == "absolute_angular":
        distance = ((1 - abs(dependence)).round(5) / 2.) ** 0.5
    return distance


def calculate_kullback_leibler_divergence(
    p: np.ndarray,
    q: np.ndarray
) -> float:
    """
    Calculates Kullback-Leibler divergence from two discrete probability distributions defined on the same probability space.

    :param p: First distribution.
    :param q: Second distribution.

    :return: Kullback-Leibler divergence.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    """
    divergence = -(p * np.log(q / p)).sum()
    return divergence


def calculate_cross_entropy(
    p: np.ndarray,
    q: np.ndarray
) -> float:
    """
    Calculates cross-entropy from two discrete probability distributions defined on the same probability space.

    :param p: First distribution.
    :param q: Second distribution.

    :return: Cross-entropy.

    Reference: De Prado, M. (2020) Advances in financial machine learning. John Wiley & Sons.
    """
    entropy = -(p * np.log(q)).sum()
    return entropy



import torch
import torch.nn as nn
import torch.autograd as autograd
import numpy as np
from .model import*

def initialize_weights(m: nn.Module) -> None:
    """
    Initializes the weights of the given module.

    Args:
    - m (nn.Module): the module to initialize weights of

    Returns:
    - None
    """
    if isinstance(m, nn.Linear):
        torch.nn.init.normal_(m.weight, 0, 0.1)
        m.bias.data.fill_(0.00)


class FBSDESolver:
    def __init__(
            self,
            pde,
            layer_sizes,
            learning_rate,
            solving_method
    ):
        """
        Initializes the FBSDESolver.

        Args:
        - pde : the partial differential equation to solve
        - layer_sizes (list[int]): list of sizes of hidden layers
        - learning_rate (float): learning rate for optimization
        - solving_method (str): method to solve the PDE ('Monte-Carlo', 'Deep-Time-SetTransformer', 'Basic')
        """
        self.pde = pde
        self.layer_size = layer_sizes
        self.learning_rate = learning_rate
        self.method = solving_method
        if solving_method == 'Monte-Carlo':
            self.solver = TimeDependentNetworkMonteCarlo(pde.dim, self.layer_size, pde.dim, pde.sigma)
        elif solving_method == 'Deep-Time-SetTransformer':
            self.solver = DeepTimeSetTransformer(1)
        elif solving_method == 'Basic':
            self.solver = TimeDependentNetwork(pde.dim, self.layer_size, pde.dim)

        self.optimizer = torch.optim.Adam(self.solver.parameters(), lr=self.learning_rate)

    def solve(
            self,
            num_iterations,
            batch_size,
            init,
            device,
            sample_size=None
    ):
        """
        Solves the PDE.

        Args:
        - num_iterations (int): number of iterations for optimization
        - batch_size (int): batch size for training
        - init (torch.Tensor): initial value
        - device (torch.device): device to perform calculations on ('cpu', 'cuda')
        - sample_size (int, optional): sample size for Monte-Carlo method

        Returns:
        - list[torch.Tensor]: list of losses during optimization
        - list[torch.Tensor]: list of initial values during optimization
        """
        losses = []
        inits = []
        self.solver.to(device)

        torch.seed()
        self.solver.train()

        for j in range(num_iterations):
            print(j + 1)

            dw, y = self.pde.sample(batch_size)
            y = torch.tensor(y)
            y = y.to(torch.float32)
            dw = torch.tensor(dw).to(torch.float32)
            dw = dw.to(device)
            t = torch.ones((batch_size, 1)).to(device)

            self.optimizer.zero_grad()

            y = y.to(device)

            y_terminal = torch.ones((batch_size,)).to(device) * init.to(device)

            y_terminal = y_terminal.to(device)
            coef = torch.ones((batch_size,)).to(device)
            dw_coef = torch.zeros((batch_size,)).to(device)
            L1 = 0.0
            if self.method == 'Monte-Carlo':
                num_sample = 5000
                primary_sample = torch.unsqueeze(torch.randn(size=[num_sample, self.pde.dim]), dim=0).to(device)
            for z in range(self.pde.num_time_interval):
                S0 = y[:, :, z]

                t0 = t * self.pde.delta_t * z
                if self.method == 'Monte-Carlo':
                    ders = []
                    time_length = (self.pde.num_time_interval - z) * self.pde.delta_t
                    current_state = torch.unsqueeze(S0, dim=1).to(device)
                    dw_sample = primary_sample * torch.tensor(np.sqrt(time_length)).to(device)
                    k = (1 + self.pde.mu_bar * time_length) * current_state + (self.pde.sigma * current_state * dw_sample)
                    ddw = dw_sample
                    means = torch.mean((self.pde.terminal_for_sample(k)), dim=1, keepdim=True)
                    mins = torch.mean((self.pde.terminal_for_sample(k) - means) * ddw, dim=1) / torch.tensor(
                        (self.pde.num_time_interval - z) * self.pde.delta_t).to(device)
                    ders = torch.tensor(mins).to(torch.float32).to(device)
                    out1 = self.solver(t0, S0, ders)
                elif self.method == 'Deep-Time-SetTransformer':
                    t = torch.ones((batch_size, 1, 1)).to(device)
                    t0 = t * self.pde.delta_t * z
                    S0 = S0.reshape(S0.size()[0], 100, 1).requires_grad_(True)
                    out1 = self.solver(t0, S0)
                    out1 = torch.squeeze(autograd.grad(out1.sum(), S0, create_graph=True)[0])

                    S0 = torch.squeeze(S0)

                    out1 = out1 * S0 * self.pde.sigma
                elif self.method == 'Basic':
                    out1 = self.solver(t0, S0)
                else:
                    print('Model type is not true')

                samp = dw[:, :, z].to(device)

                nsim = S0.size()[0]
                interest_rate = torch.squeeze(self.pde.r_u(t0, S0, y_terminal, out1).to(device))
                hz = self.pde.h_z(t0, S0, y_terminal, out1)
                if z > 0:
                    dw_coef = dw_coef * (1 + interest_rate * self.pde.delta_t) + torch.squeeze(
                        torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2))) + torch.squeeze(
                        self.pde.h_z(t0, S0, y_terminal, out1)).to(device) * self.pde.delta_t
                else:
                    dw_coef = torch.squeeze(
                        torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2))) + torch.squeeze(
                        self.pde.h_z(t0, S0, y_terminal, out1)).to(device) * self.pde.delta_t
                coef = coef * (1 + interest_rate * self.pde.delta_t)
                y_terminal = y_terminal * (1 + interest_rate * self.pde.delta_t) + torch.squeeze(
                    self.pde.h_z(t0, S0, y_terminal, out1).to(device)) * self.pde.delta_t + torch.squeeze(
                    torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2)))

            S0 = y[:, :, self.pde.num_time_interval]
            t0 = t * self.pde.delta_t * self.pde.num_time_interval

            payoff = self.pde.terminal(t0, S0)

            coef = coef.to(device)
            inits.append(init)
            init = (torch.mean((torch.squeeze(payoff) - dw_coef.detach()) / coef).detach())
            print(init)

            loss = torch.mean(torch.square(torch.squeeze(payoff) - torch.squeeze(y_terminal)))

            losses.append(loss.detach())

            print(loss)

            loss.backward()

            self.optimizer.step()

        return losses, inits



from typing import Tuple, Union

import numpy as np

from RiskLabAI.data.structures.abstract_imbalance_bars import AbstractImbalanceBars
from RiskLabAI.utils.constants import *
from RiskLabAI.utils.ewma import ewma


class ExpectedImbalanceBars(AbstractImbalanceBars):
    """
    Concrete class that contains the properties which are shared between all various type of ewma imbalance bars (dollar, volume, tick).
    """

    def __init__(
            self,
            bar_type: str,
            window_size_for_expected_n_ticks_estimation: int,
            initial_estimate_of_expected_n_ticks_in_bar: int,
            window_size_for_expected_imbalance_estimation: int,
            expected_ticks_number_bounds: Tuple[float, float],
            analyse_thresholds: bool
    ):
        """
        ExpectedImbalanceBars constructor function
        :param bar_type: type of bar. e.g. expected_dollar_imbalance_bars, fixed_tick_imbalance_bars etc.
        :param window_size_for_expected_n_ticks_estimation: window size used to estimate number of ticks expectation
        :param initial_estimate_of_expected_n_ticks_in_bar: initial estimate of number of ticks expectation window size
        :param window_size_for_expected_imbalance_estimation: window size used to estimate imbalance expectation
        :param expected_ticks_number_bounds lower and upper bound of possible number of expected ticks that used to force bars sampling convergence.
        :param analyse_thresholds: whether return thresholds values (θ, number of ticks expectation, imbalance expectation) in a tabular format
        """

        AbstractImbalanceBars.__init__(
            self,
            bar_type,
            window_size_for_expected_n_ticks_estimation,
            window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds
        )

        if expected_ticks_number_bounds is None:
            self.expected_ticks_number_lower_bound, self.expected_ticks_number_upper_bound = 0, np.inf
        else:
            self.expected_ticks_number_lower_bound, self.expected_ticks_number_upper_bound = expected_ticks_number_bounds

    def _expected_number_of_ticks(self) -> Union[float, int]:
        """
        Calculate number of ticks expectation when new imbalance bar is sampled.

        :return: number of ticks expectation.
        """

        previous_bars_n_ticks_list = self.imbalance_bars_statistics[PREVIOUS_BARS_N_TICKS_LIST]
        expected_ticks_number = ewma(np.array(
            previous_bars_n_ticks_list[-self.window_size_for_expected_n_ticks_estimation:], dtype=float),
            self.window_size_for_expected_n_ticks_estimation)[-1]

        return min(max(expected_ticks_number, self.expected_ticks_number_lower_bound),
                   self.expected_ticks_number_upper_bound)


class FixedImbalanceBars(AbstractImbalanceBars):
    """
    Concrete class that contains the properties which are shared between all various type of const imbalance bars (dollar, volume, tick).
    """

    def __init__(
            self,
            bar_type: str,
            window_size_for_expected_n_ticks_estimation: int,
            initial_estimate_of_expected_n_ticks_in_bar: int,
            window_size_for_expected_imbalance_estimation: int,
            analyse_thresholds: bool
    ):
        """
        FixedImbalanceBars constructor function
        :param bar_type: type of bar. e.g. expected_dollar_imbalance_bars, fixed_tick_imbalance_bars etc.
        :param window_size_for_expected_n_ticks_estimation: window size used to estimate number of ticks expectation
        :param initial_estimate_of_expected_n_ticks_in_bar: initial estimate of number of ticks expectation window size
        :param window_size_for_expected_imbalance_estimation: window size used to estimate imbalance expectation
        :param analyse_thresholds: whether return thresholds values (θ, number of ticks expectation, imbalance expectation) in a tabular format
        """

        AbstractImbalanceBars.__init__(self, bar_type, window_size_for_expected_n_ticks_estimation,
                                       window_size_for_expected_imbalance_estimation,
                                       initial_estimate_of_expected_n_ticks_in_bar, analyse_thresholds)

    def _expected_number_of_ticks(self) -> Union[float, int]:
        """
        Calculate number of ticks expectation when new imbalance bar is sampled.

        :return: number of ticks expectation.
        """

        return self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER]



from .kfold import KFold
from .walk_forward import WalkForward
from .purged_kfold import PurgedKFold
from .combinatorial_purged import CombinatorialPurged
from .cross_validator_interface import CrossValidator

class CrossValidatorFactory:
    """
    Factory class for creating cross-validator objects.
    """

    @staticmethod
    def create_cross_validator(
            validator_type: str,
            **kwargs
    ) -> CrossValidator:
        """
        Factory method to create and return an instance of a cross-validator
        based on the provided type.

        :param validator_type: Type of cross-validator to create. Options include
            'kfold', 'walkforward', 'purgedkfold', and 'combinatorialpurged'.
        :type validator_type: str

        :param kwargs: Additional keyword arguments to be passed to the cross-validator's constructor.
        :type kwargs: Type

        :return: An instance of the specified cross-validator.
        :rtype: CrossValidator

        :raises ValueError: If an invalid validator type is provided.
        """
        if validator_type == 'kfold':
            return KFold(**kwargs)
            
        elif validator_type == 'walkforward':
            return WalkForward(**kwargs)

        elif validator_type == 'purgedkfold':
            return PurgedKFold(**kwargs)

        elif validator_type == 'combinatorialpurged':
            return CombinatorialPurged(**kwargs)

        else:
            raise ValueError(f"Invalid validator_type: {validator_type}")
